@@ -26,7 +26,7 @@ use crate::{
     circuit_node_auto_impl, circuit_node_extra_impl,
     circuit_node_private::{CircuitNodeHashItems, CircuitNodePrivate, CircuitNodeSetNonHashInfo},
     circuit_utils::OperatorPriority,
-    deep_map, deep_map_op_context_preorder_stoppable, deep_map_preorder_unwrap, deep_map_unwrap,
+    deep_map, deep_map_op_context_preorder_stoppable, deep_map_preorder_unwrap,
     expand_node::{ExpandError, ReplaceExpander, ReplaceMapRc},
     named_axes::{deep_strip_axis_names, propagate_named_axes},
     new_rc_unwrap,
@@ -191,24 +191,18 @@ impl ModuleSpec {
         Ok(out)
     }
 
-    pub fn expand_shape(&self, shapes: &Vec<Shape>) -> Result<(CircuitRc, Vec<HashBytes>)> {
+    pub fn expand_shape(&self, shapes: &Vec<Shape>) -> Result<(CircuitRc, Vec<HashBytes>, String)> {
         let key = (self.get_hashable(), shapes.clone());
         if let Some(result) = MODULE_EXPANSIONS_SHAPE.with(|cache| {
             let borrowed = cache.borrow();
-            if let Some((w, shapes)) = borrowed.get(&key) {
-                return Some((w.clone(), shapes.clone()));
-
-                // if let Some(w) = w.upgrade() {
-                //     return Some((CircuitRc(w), shapes.clone()));
-                // } else {
-                //     drop(borrowed);
-                //     cache.borrow_mut().remove(&key);
-                // }
+            if let Some((w, shapes, n)) = borrowed.get(&key) {
+                return Some((w.clone(), shapes.clone(), n.clone()));
             }
             None
         }) {
             return Ok(result);
         }
+        let uuid = format!("{}", Uuid::new_v4());
         let symbols = shapes
             .iter()
             .enumerate()
@@ -216,7 +210,12 @@ impl ModuleSpec {
                 Symbol::nrc(
                     s.clone(),
                     self.get_spec_circuit_uuid(),
-                    Some(format!("{}_{:?}", i, s)),
+                    Some(format!(
+                        "{}_internal_expand_shape_{}_arg{}",
+                        uuid,
+                        self.circuit.name().unwrap_or(""),
+                        i
+                    )),
                 )
             })
             .collect();
@@ -225,14 +224,10 @@ impl ModuleSpec {
         MODULE_EXPANSIONS_SHAPE.with(|cache| {
             cache.borrow_mut().insert(
                 (self.get_hashable(), shapes.clone()),
-                (
-                    // Arc::downgrade(&result.0)
-                    result.clone(),
-                    out_bytes.clone(),
-                ),
+                (result.clone(), out_bytes.clone(), uuid.clone()),
             )
         });
-        Ok((result, out_bytes))
+        Ok((result, out_bytes, uuid))
     }
 
     /// we could check this on module spec construct if we wanted to
@@ -317,29 +312,30 @@ impl ModuleSpec {
         self.check_no_free_bound_by_nested(nodes)?;
 
         let shapes = nodes.iter().map(|x| x.info().shape.clone()).collect();
-        let (mut expanded_shape, rep_hashes) = self.expand_shape(&shapes)?;
+        let (expanded_shape, rep_hashes, clear_names_str) = self.expand_shape(&shapes)?;
         let node_mapping: HashMap<_, _> = rep_hashes.into_iter().zip(nodes).collect();
-        if name_prefix.is_some() || name_suffix.is_some() {
-            expanded_shape = deep_map_unwrap(expanded_shape, |x| {
+        // we're replacing leaves so no need to do anything with stoppable preorder
+        let result = deep_map(expanded_shape, |x| {
+            if let Some(r) = node_mapping.get(&x.info().hash) {
+                Ok((*r).clone())
+            } else {
+                let mut y = x.clone();
                 if let Some(n) = x.name() {
-                    if !node_mapping.contains_key(&x.info().hash) {
-                        return x.clone().rename(Some(
-                            name_prefix.clone().unwrap_or_else(|| "".to_owned())
-                                + n
-                                + name_suffix.as_ref().map(|x| -> &str { x }).unwrap_or(""),
-                        ));
+                    if n.contains(&clear_names_str) {
+                        y = y.rename(x.get_autoname().unwrap_or(None));
+                    }
+                    if name_prefix.is_some() || name_suffix.is_some() {
+                        if !node_mapping.contains_key(&x.info().hash) {
+                            y = y.rename(Some(
+                                name_prefix.clone().unwrap_or_else(|| "".to_owned())
+                                    + n
+                                    + name_suffix.as_ref().map(|x| -> &str { x }).unwrap_or(""),
+                            ));
+                        }
                     }
                 }
-                x.clone()
-            })
-        }
-        // we're replacing leaves so no need to do anything with stoppable preorder
-        let result = deep_map(expanded_shape, |circ| {
-            Ok(node_mapping
-                .get(&circ.info().hash)
-                .cloned()
-                .cloned()
-                .unwrap_or(circ))
+                Ok(y)
+            }
         })
         .with_context(|| {
             format!(
@@ -1181,7 +1177,7 @@ pub fn has_free_sym(circuit: &CircuitRc, sym: &Symbol) -> bool {
     }
 }
 
-pub fn intersect_free_syms(circuit: &CircuitRc, syms: &[&Symbol]) -> bool {
+pub fn is_intersecting_free_syms_slice(circuit: &CircuitRc, syms: &[&Symbol]) -> bool {
     if let Some(other) = circuit.as_symbol() {
         syms.contains(&other)
     } else {
@@ -1190,6 +1186,14 @@ pub fn intersect_free_syms(circuit: &CircuitRc, syms: &[&Symbol]) -> bool {
     }
 }
 
+pub fn is_intersecting_free_syms(circuit: &CircuitRc, syms: &IndexSet<Symbol>) -> bool {
+    if let Some(other) = circuit.as_symbol() {
+        syms.contains(other)
+    } else {
+        !circuit.info().free_symbols.is_disjoint(syms)
+    }
+}
+
 pub fn recur_check_no_free_bound_by_nested(
     circ: CircuitRc,
     syms: &[&Symbol],
@@ -1199,7 +1203,7 @@ pub fn recur_check_no_free_bound_by_nested(
     if seen.insert(circ.info().hash) {
         if let Some(m) = circ.as_module() {
             // first check if we can term early (maybe not actually optimization)
-            if !intersect_free_syms(&circ, syms) {
+            if !is_intersecting_free_syms_slice(&circ, syms) {
                 return Ok(());
             }
 
@@ -1217,7 +1221,7 @@ pub fn recur_check_no_free_bound_by_nested(
                 })
                 .unzip();
 
-            if intersect_free_syms(&m.spec.circuit, &syms) {
+            if is_intersecting_free_syms_slice(&m.spec.circuit, &syms) {
                 // fail earlier if possible
                 recur_check_no_free_bound_by_nested(m.spec.circuit.clone(), &syms, &nodes, seen)?;
 
@@ -1289,7 +1293,7 @@ thread_local! {
     static MODULE_EXPANSIONS_SHAPE: RefCell<
         HashMap<
             (ModuleSpecHashable, Vec<Shape>),
-            (CircuitRc, Vec<HashBytes>),
+            (CircuitRc, Vec<HashBytes>, String),
         >,
     > = RefCell::new(HashMap::default());
     static MODULE_EXPANSIONS_AXIS_MAPS: RefCell<
