@@ -14,7 +14,9 @@ use pyo3::{exceptions::PyTypeError, prelude::*, AsPyPointer};
 use regex::{Captures, Regex};
 use rr_util::{
     eq_by_big_hash::EqByBigHash,
-    make_single_many, setup_callable, simple_default, simple_from,
+    make_single_many,
+    py_types::PyCallable,
+    setup_callable, simple_default, simple_from,
     util::{arc_unwrap_or_clone, EmptySingleMany as ESM},
 };
 use thiserror::Error;
@@ -41,7 +43,7 @@ pub enum MatcherFromPyBase {
 pub enum MatcherFromPy {
     Base(MatcherFromPyBase),
     #[pyo3(transparent)]
-    PyFunc(PyObject),
+    PyFunc(PyCallable),
 }
 
 #[derive(Clone, Debug)]
@@ -53,7 +55,7 @@ pub enum MatcherData {
     EqM(BTreeSet<CircuitRc>),
     AnyFound(Arc<Mutex<BoundAnyFound>>), // sad mutex
     Raw(RawMatcher),
-    PyFunc(PyObject),
+    PyFunc(PyCallable),
     Not(MatcherRc),
     Any(Vec<MatcherRc>),
     All(Vec<MatcherRc>),
