@@ -8,12 +8,12 @@ use std::{
 use anyhow::{anyhow, bail, Context, Result};
 use circuit_base::{
     expand_node::{expand_node, MapReplaceExpander, ReplaceMapRc},
-    module::are_args_used,
+    module::{are_args_used, is_intersecting_free_syms},
     CircuitNode, CircuitRc, Module, ModuleArgSpec, ModuleSpec, Rearrange, Symbol,
 };
 use circuit_rewrites::{
     deep_rewrite::SimpFnSubset,
-    module_rewrite::{extract_rewrite_raw, module_remove_unused_inputs},
+    module_rewrite::{elim_no_input_module, extract_rewrite_raw, module_remove_unused_inputs},
 };
 use get_update_node::{AnyFound, IterativeMatcher, IterativeMatcherRc, MatcherData};
 use itertools::multiunzip;
@@ -29,8 +29,9 @@ use rr_util::{
     eq_by_big_hash::EqByBigHash,
     fn_struct, impl_eq_by_big_hash, pycall, python_error_exception,
     rearrange_spec::{OpSize, RearrangeSpec, RearrangeSpecError},
-    sv, tu8v,
+    sv, tu8v, unwrap,
     util::{arc_unwrap_or_clone, DimNumMaker, HashBytes},
+    IndexSet,
 };
 use rustc_hash::{FxHashMap as HashMap, FxHashSet as HashSet};
 use thiserror::Error;
@@ -395,52 +396,54 @@ fn_struct!(
     pub NestedModuleNamer: Fn(
         base_circuit: CircuitRc,
         running_circuit: CircuitRc,
-        modules: Vec<Module>
+        modules: Vec<Module>,
+        pushed_overall_mod_count: Option<usize>
     ) -> Option<String>
 );
 
-#[pyfunction]
-pub fn default_nested_module_namer() -> NestedModuleNamer {
-    Default::default()
+#[pyfunction(bind_name = "\"bind\".to_owned()")]
+pub fn default_nested_module_namer(bind_name: String) -> NestedModuleNamer {
+    NestedModuleNamer::Dyn(NestedModuleNamerDynStruct(Arc::new(
+        move |base_circuit, _, modules, _| {
+            if modules.is_empty() {
+                return Ok(base_circuit.name_cloned());
+            }
+
+            if base_circuit.name().is_none() || modules.iter().any(|m| m.name().is_none()) {
+                return Ok(None);
+            }
+
+            Ok(Some(format!(
+                "{} {bind_name}:{}",
+                base_circuit.name().unwrap(),
+                modules
+                    .iter()
+                    .map(|x| x.name().unwrap())
+                    .collect::<Vec<_>>()
+                    .join(",")
+            )))
+        },
+    )))
 }
 
 impl Default for NestedModuleNamer {
     fn default() -> Self {
-        NestedModuleNamer::Dyn(NestedModuleNamerDynStruct(std::sync::Arc::new(
-            |base_circuit, _, modules| {
-                if modules.is_empty() {
-                    return Ok(base_circuit.name_cloned());
-                }
-
-                if base_circuit.name().is_none() || modules.iter().any(|m| m.name().is_none()) {
-                    return Ok(None);
-                }
-
-                Ok(Some(format!(
-                    "{} bind:{}",
-                    base_circuit.name().unwrap(),
-                    modules
-                        .iter()
-                        .map(|x| x.name().unwrap())
-                        .collect::<Vec<_>>()
-                        .join(",")
-                )))
-            },
-        )))
+        default_nested_module_namer("bind".to_owned())
     }
 }
 
+fn_struct!(pub ModuleConstructCallback: FnMut(m: Module, applied_modules: Vec<Module>, pushed_overall_mod_count: usize) -> CircuitRc;
+{
+    UpdateBindings(ModuleConstructUpdateBindings),
+});
+
 #[pyclass]
 #[derive(Debug, Clone)]
 pub struct ModulePusher {
     #[pyo3(get)]
     flatten_modules: bool,
     #[pyo3(get)]
-    remove_unused_inputs: bool,
-    #[pyo3(get)]
-    add_suffix_on_remove_unused: bool,
-    #[pyo3(get)]
-    elim_no_input_modules: bool,
+    module_construct_callback: ModuleConstructCallback,
     #[pyo3(get)]
     bind_encountered_symbols: bool,
     #[pyo3(get)]
@@ -451,11 +454,11 @@ pub struct ModulePusher {
             HashBytes,
             IterativeMatcherRc,
             IterativeMatcherRc,
-            bool,
+            PushDownMode,
             HashBytes,
             HashBytes,
         ),
-        (CircuitRc, HashSet<CircuitRc>),
+        CircOrGet,
     >,
     replace_expander: MapReplaceExpander,
 }
@@ -464,9 +467,7 @@ impl Default for ModulePusher {
     fn default() -> Self {
         Self {
             flatten_modules: true,
-            remove_unused_inputs: true,
-            add_suffix_on_remove_unused: false,
-            elim_no_input_modules: true,
+            module_construct_callback: Self::remove_unused_callback(false, true),
             bind_encountered_symbols: true,
             namer: Default::default(),
             any_found: Default::default(),
@@ -476,54 +477,36 @@ impl Default for ModulePusher {
     }
 }
 
+#[derive(Clone, Debug)]
+enum CircOrGet {
+    Circ(CircuitRc),
+    Get(HashSet<CircuitRc>),
+}
+
 #[pymethods]
 impl ModulePusher {
     #[new]
     #[args(
         flatten_modules = "Self::default().flatten_modules",
-        remove_unused_inputs = "Self::default().remove_unused_inputs",
-        add_suffix_on_remove_unused = "Self::default().add_suffix_on_remove_unused",
-        elim_no_input_modules = "Self::default().elim_no_input_modules",
+        module_construct_callback = "Self::default().module_construct_callback",
         bind_encountered_symbols = "Self::default().bind_encountered_symbols",
         namer = "Self::default().namer"
     )]
     pub fn new(
         flatten_modules: bool,
-        remove_unused_inputs: bool,
-        add_suffix_on_remove_unused: bool,
-        elim_no_input_modules: bool,
+        module_construct_callback: ModuleConstructCallback,
         bind_encountered_symbols: bool,
         namer: NestedModuleNamer,
     ) -> Self {
         Self {
             flatten_modules,
-            remove_unused_inputs,
-            add_suffix_on_remove_unused,
-            elim_no_input_modules,
+            module_construct_callback,
             bind_encountered_symbols,
             namer,
             ..Default::default()
         }
     }
 
-    #[args(skip_module = "MatcherData::Always(false).into()", is_get = "true")]
-    pub fn push_down_modules_get(
-        &mut self,
-        circuit: CircuitRc,
-        get: IterativeMatcherRc,
-        skip_module: IterativeMatcherRc,
-        is_get: bool,
-    ) -> Result<(CircuitRc, HashSet<CircuitRc>)> {
-        self.push_down_modules_rec(
-            circuit,
-            get,
-            skip_module,
-            is_get,
-            &Default::default(),
-            &Default::default(),
-        )
-    }
-
     #[args(skip_module = "MatcherData::Always(false).into()")]
     fn __call__(
         &mut self,
@@ -542,8 +525,8 @@ impl ModulePusher {
         traversal: IterativeMatcherRc,
         skip_module: IterativeMatcherRc,
     ) -> Result<CircuitRc> {
-        self.push_down_modules_get(circuit, traversal, skip_module, false)
-            .map(|x| x.0)
+        self.push_down_modules_impl(circuit, traversal, skip_module, PushDownMode::Circ)
+            .map(|x| unwrap!(x, CircOrGet::Circ))
     }
 
     // TODO: fancy validate
@@ -554,8 +537,8 @@ impl ModulePusher {
         get: IterativeMatcherRc,
         skip_module: IterativeMatcherRc,
     ) -> Result<HashSet<CircuitRc>> {
-        self.push_down_modules_get(circuit, get, skip_module, true)
-            .map(|x| x.1)
+        self.push_down_modules_impl(circuit, get, skip_module, PushDownMode::Get)
+            .map(|x| unwrap!(x, CircOrGet::Get))
     }
 
     #[args(skip_module = "MatcherData::Always(false).into()")]
@@ -582,12 +565,91 @@ impl ModulePusher {
         self.get_unique_op_push_down_modules(circuit, get, skip_module)?
             .ok_or_else(|| anyhow!("found no matches!"))
     }
+
+    #[staticmethod]
+    #[args(
+        remove_unused_inputs = "true",
+        add_suffix_on_remove_unused = "false",
+        elim_no_input_modules = "true"
+    )]
+    pub fn remove_and_elim_callback(
+        remove_unused_inputs: bool,
+        add_suffix_on_remove_unused: bool,
+        elim_no_input_modules: bool,
+    ) -> ModuleConstructCallback {
+        if !remove_unused_inputs && !elim_no_input_modules {
+            Self::noop_callback()
+        } else if !remove_unused_inputs {
+            Self::elim_no_input_modules_callback()
+        } else {
+            Self::remove_unused_callback(add_suffix_on_remove_unused, elim_no_input_modules)
+        }
+    }
+
+    #[staticmethod]
+    #[args(add_suffix_on_remove_unused = "false", elim_no_input_modules = "true")]
+    pub fn remove_unused_callback(
+        add_suffix_on_remove_unused: bool,
+        elim_no_input_modules: bool,
+    ) -> ModuleConstructCallback {
+        ModuleConstructCallback::Dyn(ModuleConstructCallbackDynStruct(Arc::new(
+            move |m, _, _| {
+                module_remove_unused_inputs(&m, add_suffix_on_remove_unused, elim_no_input_modules)
+            },
+        )))
+    }
+
+    #[staticmethod]
+    pub fn noop_callback() -> ModuleConstructCallback {
+        ModuleConstructCallback::Dyn(ModuleConstructCallbackDynStruct(Arc::new(
+            move |m, _, _| Ok(m.rc()),
+        )))
+    }
+
+    #[staticmethod]
+    pub fn elim_no_input_modules_callback() -> ModuleConstructCallback {
+        ModuleConstructCallback::Dyn(ModuleConstructCallbackDynStruct(Arc::new(
+            move |m, _, _| Ok(elim_no_input_module(&m).unwrap_or_else(|| m.rc())),
+        )))
+    }
+}
+
+#[derive(Debug, Clone, Copy, PartialEq, Eq, Hash)]
+enum PushDownMode {
+    PushOverride,
+    Circ,
+    Get,
 }
 
 impl ModulePusher {
+    fn push_down_modules_impl(
+        &mut self,
+        circuit: CircuitRc,
+        get: IterativeMatcherRc,
+        skip_module: IterativeMatcherRc,
+        mode: PushDownMode,
+    ) -> Result<CircOrGet> {
+        self.push_down_modules_rec(
+            circuit,
+            get,
+            skip_module,
+            mode,
+            &Default::default(),
+            &Default::default(),
+        )
+    }
+
+    // we could do a generic function on the bools, but I assume this has better compile time
     #[apply(cached_method)]
     #[self_id(self_)]
-    #[key((circuit.info().hash,  get.clone(), skip_module.clone(), is_get, items.hash(), extra_replacements.hash()))]
+    #[key((
+        circuit.info().hash,
+        get.clone(),
+        skip_module.clone(),
+        mode,
+        items.hash(),
+        extra_replacements.hash(),
+    ))]
     #[use_try]
     #[cache_expr(cache)]
     fn push_down_modules_rec(
@@ -595,66 +657,87 @@ impl ModulePusher {
         circuit: CircuitRc,
         get: IterativeMatcherRc,
         skip_module: IterativeMatcherRc,
-        is_get: bool,
+        mode: PushDownMode,
         items: &NestedModuleItems,
         extra_replacements: &ReplaceMapRc,
-    ) -> Result<(CircuitRc, HashSet<CircuitRc>)> {
+    ) -> Result<CircOrGet> {
+        let push_or_get = mode == PushDownMode::PushOverride || mode == PushDownMode::Get;
+        if push_or_get {
+            assert_eq!(extra_replacements.len(), 0);
+        }
         let (new_get, found_get) = get
             .match_iterate(circuit.clone())?
             .unwrap_or_same(get.clone());
-        let all_done = new_get.all_finished()
-            || (is_get && !self_.any_found.are_any_found(circuit.clone(), get)?);
+
+        let is_get = mode == PushDownMode::Get;
+        let all_finished = new_get.all_finished();
+        let new_get = new_get.clone().per_child_with_term(circuit.num_children());
+        // before get_fin for borrow check reasons
+        let no_child_will_get = push_or_get
+            && !circuit
+                .children()
+                .zip(&new_get)
+                .map(|(c, get_per)| self_.any_found.are_any_found(c, get_per.clone()))
+                .collect::<Result<Vec<_>>>()?
+                .into_iter()
+                .any(|x| x);
 
         let mut get_fin = || self_.finalize(circuit.clone(), items, extra_replacements);
         let mut out_set = HashSet::default();
 
-        let finalized = if is_get && found_get {
+        if is_get && found_get {
             let out = get_fin()?;
             out_set.insert(out.clone());
-            Some(out)
-        } else {
-            None
-        };
+        }
+
+        let found_override = found_get && mode == PushDownMode::PushOverride;
+        let all_done = all_finished || found_override || no_child_will_get;
 
         if all_done {
-            let out = if let Some(fin) = finalized {
-                // avoid recomputation
-                fin
-            } else {
-                get_fin()?
+            let out = match (found_override, mode) {
+                (_, PushDownMode::Get) => CircOrGet::Get(out_set),
+                (_, PushDownMode::Circ) | (true, PushDownMode::PushOverride) => {
+                    CircOrGet::Circ(get_fin()?)
+                }
+                (false, PushDownMode::PushOverride) => CircOrGet::Circ(circuit),
             };
-            return Ok((out, out_set));
+            return Ok(out);
         }
 
         if let Some(arg) = circuit
             .as_symbol()
             .and_then(|sym| items.flat_sym_to_arg.get(sym))
         {
+            assert!(!push_or_get);
             if !self_.bind_encountered_symbols {
                 bail!(PushDownModuleError::PushPastPreviouslyBoundSymbol {
                     symbol: circuit.as_symbol().unwrap().clone()
                 })
             }
-            return Ok((arg.clone(), Default::default()));
+
+            return Ok(CircOrGet::Circ(arg.clone()));
         }
         if let Some(out) = extra_replacements.get(&circuit) {
-            return Ok((out.clone(), Default::default()));
+            return Ok(CircOrGet::Circ(out.clone()));
         }
 
-        let new_get = new_get.per_child_with_term(circuit.num_children());
-
         let (new_skip_module, found_skip_module) = skip_module
             .match_iterate(circuit.clone())?
             .unwrap_or_same(skip_module);
         let new_skip_module = new_skip_module.per_child_with_term(circuit.num_children());
 
-        let mut rec = |self_: &mut Self, c, get, skip, items, extra_replacements: &_| {
-            let (out, new_set) =
-                self_.push_down_modules_rec(c, get, skip, is_get, items, extra_replacements)?;
-            if is_get {
+        let mut rec = |self_: &mut Self, c, get, skip, items, extra_replacements: &_| match self_
+            .push_down_modules_rec(c, get, skip, mode, items, extra_replacements)?
+        {
+            CircOrGet::Circ(c) => {
+                assert!(!is_get);
+                Ok(Some(c))
+            }
+            CircOrGet::Get(new_set) => {
+                assert!(is_get);
                 out_set.extend(new_set);
+                Ok(None)
             }
-            Ok(out)
         };
 
         let out = if let Some(m) = circuit
@@ -667,7 +750,7 @@ impl ModulePusher {
             assert_eq!(node_get_skip.len(), m.nodes.len());
 
             // intentionally don't use extra_replacements per child
-            let new_nodes_flat = m
+            let new_nodes_flat_orig = m
                 .nodes
                 .clone()
                 .into_iter()
@@ -675,36 +758,83 @@ impl ModulePusher {
                 .map(|(c, (get, skip))| rec(self_, c, get, skip, items, extra_replacements))
                 .collect::<Result<Vec<_>>>()?;
 
-            let new_nodes_base = m
-                .nodes
-                .clone()
-                .into_iter()
-                .zip(node_get_skip)
-                // don't use rec to avoid adding nodes to out_set
-                .map(|(c, (get, skip))| {
-                    let ndim = c.ndim();
-                    let out = self_
-                        .push_down_modules_rec(
+            let (new_nodes_flat, new_nodes_base): (Vec<_>, _) = if push_or_get {
+                assert_eq!(extra_replacements.len(), 0);
+                let new_nodes_base = m.nodes.clone();
+
+                // we fully substitute in the 'get' case
+
+                (
+                    m.nodes
+                        .iter()
+                        .cloned()
+                        .map(|circuit| {
+                            let out = self_
+                                .build_module_flat_no_call(
+                                    circuit.clone(),
+                                    circuit.clone(),
+                                    items.flat.clone(),
+                                    items.modules.clone().into_iter().rev().collect(),
+                                    items.modules.len(),
+                                )?
+                                .substitute(None, None);
+                            Ok(out)
+                        })
+                        .collect::<Result<_>>()?,
+                    new_nodes_base,
+                )
+            } else {
+                let new_nodes_base = m
+                    .nodes
+                    .clone()
+                    .into_iter()
+                    .zip(node_get_skip)
+                    // don't use rec to avoid adding nodes to out_set
+                    .map(|(c, (get, skip))| {
+                        let ndim = c.ndim();
+                        let out = unwrap!(self_.push_down_modules_rec(
                             c,
                             get,
                             skip,
-                            is_get,
+                            mode,
                             &Default::default(), // no items (only difference from recursive call above)
                             extra_replacements,
-                        )?.0;
-                    assert_eq!(out.ndim(), ndim);
-                    Ok(out)
-                })
-                .collect::<Result<Vec<_>>>()?;
+                        )?, CircOrGet::Circ);
+                        assert_eq!(out.ndim(), ndim);
+                        Ok(out)
+                    })
+                    .collect::<Result<Vec<_>>>()?;
+                (
+                    new_nodes_flat_orig
+                        .iter()
+                        .map(|x| x.clone().unwrap())
+                        .collect(),
+                    new_nodes_base,
+                )
+            };
 
-            rec(
+            let out = rec(
                 self_,
                 m.spec.circuit.clone(),
                 spec_circuit_get,
                 spec_circuit_skip_module,
-                &items.push(m.clone(), new_nodes_flat, new_nodes_base)?,
+                &items.push(m.clone(), new_nodes_flat.clone(), new_nodes_base)?,
                 extra_replacements,
-            )?
+            )?;
+            if mode == PushDownMode::PushOverride {
+                Some(
+                    m.map_children_unwrap_idxs(|i| {
+                        if i == 0 {
+                            out.clone().unwrap()
+                        } else {
+                            new_nodes_flat_orig[i - 1].clone().unwrap()
+                        }
+                    })
+                    .rc(),
+                )
+            } else {
+                out
+            }
         } else {
             if let Some(m) = circuit.as_module() {
                 for arg_spec in &m.spec.arg_specs {
@@ -727,23 +857,38 @@ impl ModulePusher {
                 .map(|(((c, get), skip), rep)| rec(self_, c, get, skip, items, &rep))
                 .collect::<Result<Vec<_>>>()?;
 
-            expand_node(circuit.clone(), &new_children, |c, rep, child_idx| {
-                rec(
-                    self_,
-                    c,
-                    new_get[child_idx].clone(),
-                    new_skip_module[child_idx].clone(),
-                    items,
-                    &extra_replacements.extend_into(rep),
-                )
-            })
+            if is_get {
+                return Ok(CircOrGet::Get(out_set));
+            }
+
+            let out = expand_node(
+                circuit.clone(),
+                &new_children.into_iter().map(|x| x.unwrap()).collect(),
+                |c, rep, child_idx| {
+                    rec(
+                        self_,
+                        c,
+                        new_get[child_idx].clone(),
+                        new_skip_module[child_idx].clone(),
+                        items,
+                        &extra_replacements.extend_into(rep),
+                    )
+                    .map(|x| x.unwrap())
+                },
+            )
             // maybe we're supposed to just panic on rearrange rank errors?
             .context(concat!(
                 "expand fail in push down modules, should be rearrange ",
                 "overflow error (otherwise internal error)"
-            ))?
+            ))?;
+            Some(out)
         };
 
+        if is_get {
+            return Ok(CircOrGet::Get(out_set));
+        }
+        let out = out.unwrap();
+
         // rename if we didn't create new
         let out = if out != circuit {
             let new_name = self_
@@ -752,6 +897,7 @@ impl ModulePusher {
                     circuit.clone(),
                     circuit.clone(),
                     items.modules.iter().rev().cloned().collect(),
+                    None,
                 )
                 .context("namer failed in push down module")?;
             out.rename(new_name)
@@ -759,7 +905,7 @@ impl ModulePusher {
             out
         };
 
-        Ok((out, out_set))
+        Ok(CircOrGet::Circ(out))
     }
 
     fn finalize(
@@ -827,6 +973,7 @@ impl ModulePusher {
                 finished_circuit,
                 new_flat,
                 items.modules.clone().into_iter().rev().collect(),
+                items.modules.len(),
             )
         } else {
             let (finished_circuit, rev_nested) = if rep_expand_extra_dims > 0 {
@@ -878,28 +1025,48 @@ impl ModulePusher {
         }
     }
 
+    fn build_module_flat_no_call(
+        &mut self,
+        base_circuit: CircuitRc, // just for naming
+        circuit: CircuitRc,
+        flat: Vec<(CircuitRc, ModuleArgSpec)>,
+        rev_modules: Vec<Module>,
+        overall_mod_count: usize,
+    ) -> Result<Module> {
+        let (nodes, arg_specs) = flat.into_iter().unzip();
+        let name = self
+            .namer
+            .call(
+                base_circuit,
+                circuit.clone(),
+                rev_modules,
+                Some(overall_mod_count),
+            )
+            .context("namer failed in push down module")?;
+        Ok(Module::new(
+            nodes,
+            ModuleSpec { circuit, arg_specs },
+            name.clone(),
+        ))
+    }
+
     fn build_module_flat(
         &mut self,
         base_circuit: CircuitRc, // just for naming
         circuit: CircuitRc,
         flat: Vec<(CircuitRc, ModuleArgSpec)>,
         rev_modules: Vec<Module>,
+        overall_mod_count: usize,
     ) -> Result<CircuitRc> {
-        let (nodes, arg_specs) = flat.into_iter().unzip();
-        let name = self
-            .namer
-            .call(base_circuit, circuit.clone(), rev_modules)
-            .context("namer failed in push down module")?;
-        let m_out = Module::new(nodes, ModuleSpec { circuit, arg_specs }, name.clone());
-        if self.remove_unused_inputs {
-            module_remove_unused_inputs(
-                &m_out,
-                self.add_suffix_on_remove_unused,
-                self.elim_no_input_modules,
-            )
-        } else {
-            Ok(m_out.rc())
-        }
+        let out = self.build_module_flat_no_call(
+            base_circuit,
+            circuit,
+            flat,
+            rev_modules.clone(),
+            overall_mod_count,
+        )?;
+        self.module_construct_callback
+            .call(out, rev_modules, overall_mod_count)
     }
 
     fn build_module_nested(
@@ -923,12 +1090,241 @@ impl ModulePusher {
         rev_nested.into_iter().zip(cum_rev_mods).fold(
             Ok(base_circuit.clone()),
             |circuit, (arg_items, running_modules)| {
-                self.build_module_flat(base_circuit.clone(), circuit?, arg_items, running_modules)
+                self.build_module_flat(
+                    base_circuit.clone(),
+                    circuit?,
+                    arg_items,
+                    running_modules,
+                    modules.len(),
+                )
             },
         )
     }
 }
 
+fn_struct!(pub MaybeUpdate: Fn(m : CircuitRc) -> Option<CircuitRc>);
+
+#[derive(Debug, Clone)]
+pub struct ModuleConstructUpdateBindings {
+    // TODO
+    cache: FastUnboundedCache<HashBytes, Option<CircuitRc>>,
+    update: MaybeUpdate,
+    run_update_on_new_spec_circuits: bool,
+    namer: NestedModuleNamer,
+    is_flatten: bool,
+}
+
+impl IntoPy<PyObject> for ModuleConstructUpdateBindings {
+    fn into_py(self, _: Python<'_>) -> PyObject {
+        // fix as needed
+        unreachable!()
+    }
+}
+
+impl ModuleConstructUpdateBindings {
+    fn call(
+        &mut self,
+        m: Module,
+        these_applied_modules: Vec<Module>,
+        overall_count: usize,
+    ) -> Result<CircuitRc> {
+        assert!(these_applied_modules.len() <= overall_count);
+        if these_applied_modules.len() != overall_count {
+            return Ok(m.rc());
+        }
+
+        self.bind_module(
+            m,
+            overall_count,
+            Default::default(),
+            &these_applied_modules,
+            overall_count,
+        )
+        .map(|(c, _, _)| c)
+    }
+
+    fn bind_module(
+        &mut self,
+        m: Module,
+        count_mods: usize,
+        outer_updated_syms: IndexSet<Symbol>,
+        applied_modules: &[Module],
+        overall_count: usize,
+    ) -> Result<(CircuitRc, CircuitRc, Vec<Module>)> {
+        assert_eq!(applied_modules.len(), count_mods);
+        let m = if (self.is_flatten || count_mods == 1) && self.run_update_on_new_spec_circuits {
+            if let Some(new_spec) = self.deep_update(m.spec.circuit.clone())? {
+                Module::try_new(
+                    m.nodes.clone(),
+                    ModuleSpec {
+                        circuit: new_spec,
+                        arg_specs: m.spec.arg_specs.clone(),
+                    },
+                    m.name_cloned(),
+                )
+                .context(
+                    "module construction failed in update bindings after updating spec circuit",
+                )?
+            } else {
+                m
+            }
+        } else {
+            m
+        };
+
+        let used_inputs = m.spec.are_args_used();
+        let (nodes, arg_specs): (Vec<_>, Vec<_>) = m
+            .arg_items()
+            .into_iter()
+            .zip(used_inputs)
+            .map(|((node, arg_spec), used)| {
+                if used {
+                    self.deep_update(node.clone()).map(|x| {
+                        x.or_else(|| {
+                            is_intersecting_free_syms(&node, &outer_updated_syms)
+                                .then(|| node.clone())
+                        })
+                        .map(|x| (x, arg_spec))
+                    })
+                } else {
+                    Ok(None)
+                }
+            })
+            .collect::<Result<Vec<_>>>()?
+            .into_iter()
+            .filter_map(|x| x)
+            .unzip();
+
+        let (new_spec_circuit, base_circuit, prior_modules) = if count_mods > 1 && !self.is_flatten
+        {
+            self.bind_module(
+                m.spec.circuit.as_module_unwrap().clone(),
+                count_mods - 1,
+                outer_updated_syms
+                    .into_iter()
+                    .chain(arg_specs.iter().map(|x| x.symbol.clone()))
+                    .collect(),
+                &applied_modules[..applied_modules.len() - 1],
+                overall_count,
+            )?
+        } else {
+            (m.spec.circuit.clone(), m.spec.circuit.clone(), Vec::new())
+        };
+
+        let new_modules = if !self.is_flatten {
+            applied_modules.to_vec()
+        } else {
+            let mut new_modules = prior_modules.clone();
+            new_modules.push(applied_modules.last().unwrap().clone());
+            new_modules
+        };
+        let name = self.namer.call(
+            base_circuit.clone(),
+            new_spec_circuit.clone(),
+            new_modules.clone(),
+            Some(overall_count),
+        )?;
+
+        let m_out = Module::try_new(
+            nodes,
+            ModuleSpec {
+                circuit: new_spec_circuit,
+                arg_specs,
+            },
+            name,
+        )
+        .context("module construction failed in update bindings after updating nodes")?;
+
+        if m_out.aligned_batch_shape().len() > 0 {
+            bail!("batching update bindings case not supported, consider doing some conforming (TODO: better err)");
+        }
+
+        let (new_circ, prior_mods) = elim_no_input_module(&m_out)
+            .map(|x| (x, prior_modules))
+            .unwrap_or_else(|| (m_out.rc(), new_modules));
+        Ok((new_circ, base_circuit, prior_mods))
+    }
+
+    #[apply(cached_method)]
+    #[self_id(self_)]
+    #[key(circuit.info().hash)]
+    #[use_try]
+    #[cache_expr(cache)]
+    fn deep_update(&mut self, circuit: CircuitRc) -> Result<Option<CircuitRc>> {
+        // somewhat different ordering than deep_map_op
+        let out = if let Some(new) = self_.update.call(circuit.clone())? {
+            Some(new)
+        } else {
+            let new_children = circuit
+                .children()
+                .map(|c| self_.deep_update(c))
+                .collect::<Result<Vec<_>>>()?;
+            new_children.iter().any(|x| x.is_some()).then(|| {
+                circuit.map_children_unwrap_enumerate(|i, c| new_children[i].clone().unwrap_or(c))
+            })
+        };
+        Ok(out)
+    }
+}
+
+#[pyfunction(bind_name = "\"upd_bind\".to_owned()", short_if_not_leaf = "true")]
+pub fn default_update_bindings_nested_namer(
+    bind_name: String,
+    short_if_not_leaf: bool,
+) -> NestedModuleNamer {
+    let default_overall_namer = default_nested_module_namer(bind_name.clone());
+    NestedModuleNamer::Dyn(NestedModuleNamerDynStruct(Arc::new(
+        move |base_circuit, running_circuit, modules, pushed_overall_mod_count| {
+            if pushed_overall_mod_count.is_none() && short_if_not_leaf {
+                Ok(base_circuit
+                    .name()
+                    .map(|name| format!("{name} {bind_name}")))
+            } else {
+                default_overall_namer.call(
+                    base_circuit,
+                    running_circuit,
+                    modules,
+                    pushed_overall_mod_count,
+                )
+            }
+        },
+    )))
+}
+
+#[pyfunction(
+    namer = "default_update_bindings_nested_namer(\"upd_bind\".to_owned(), true)",
+    skip_module = "MatcherData::Always(false).into()",
+    run_update_on_new_spec_circuits = "false",
+    flatten_modules = "false"
+)]
+pub fn update_bindings_nested(
+    circuit: CircuitRc,
+    update: MaybeUpdate,
+    matcher: IterativeMatcherRc,
+    namer: NestedModuleNamer,
+    skip_module: IterativeMatcherRc,
+    run_update_on_new_spec_circuits: bool,
+    flatten_modules: bool,
+) -> Result<CircuitRc> {
+    let module_construct_callback =
+        ModuleConstructCallback::UpdateBindings(ModuleConstructUpdateBindings {
+            cache: Default::default(),
+            update,
+            run_update_on_new_spec_circuits,
+            namer: namer.clone(),
+            is_flatten: flatten_modules,
+        });
+
+    ModulePusher {
+        flatten_modules,
+        module_construct_callback,
+        namer,
+        ..Default::default()
+    }
+    .push_down_modules_impl(circuit, matcher, skip_module, PushDownMode::PushOverride)
+    .map(|x| unwrap!(x, CircOrGet::Circ))
+}
+
 #[derive(Clone, Debug)]
 pub struct SymbolSetRc {
     set: Arc<BTreeSet<Symbol>>,
