@@ -8,9 +8,10 @@ use pyo3::{
     pyclass::CompareOp,
     types::{IntoPyDict, PyTuple},
 };
+use rustc_hash::FxHashMap as HashMap;
 
 use crate::{
-    tensor_util::{Shape, TorchDeviceDtype},
+    tensor_util::{Shape, TorchDeviceDtype, TorchDtype},
     util::{EinsumAxes, HashBytes, HashBytesToPy},
 };
 pub struct PyUtils {
@@ -20,6 +21,7 @@ pub struct PyUtils {
     pub cast_int: PyObject,
     scalar_to_tensor: PyObject,
     pub cast_tensor: PyObject,
+    pub dtype_convert: HashMap<TorchDtype, PyObject>,
     un_flat_concat: PyObject,
     tensor_scale: PyObject,
     pub generalfunctions: std::collections::HashMap<String, PyObject>,
@@ -62,6 +64,11 @@ pub static PY_UTILS: GILLazyPy<PyUtils> = GILLazyPy::new_py(|py| {
         cast_int: get_builtin("int"),
         scalar_to_tensor: get("scalar_to_tensor"),
         cast_tensor: get("cast_tensor"),
+        dtype_convert: utils
+            .getattr("dtype_from_string")
+            .unwrap()
+            .extract()
+            .unwrap(),
         un_flat_concat: get("un_flat_concat"),
         tensor_scale: get("tensor_scale"),
         generalfunctions: utils
